/* lpmutil.c */

/*
 * lpmutil -- Moves non-QMI-enabled modems out of LPM
 * 				by using the MBIM-over-QMI service.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 * Collin McMillan <cmc@nd.edu>
 * July 31, 2015
 * 
 * SUBSTANTIAL PORTIONS OF THIS PROGRAM LIFTED SHAMELESSLY FROM MBIMCLI SOURCE
 * 
 * Details:
 * http://lists.freedesktop.org/archives/libmbim-devel/2015-July/000604.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <locale.h>
#include <string.h>
#include <errno.h>

#include <glib.h>
#include <glib/gprintf.h>
#include <gio/gio.h>

#include <libmbim-glib.h>

static gboolean verbose;

static MbimDevice *device;
static MbimService service;
static GCancellable *cancellable;
static GMainLoop *loop;

static guint32 qmi_cid;

/* These are the raw QMI messages that need to be passed, except that the 0xFF in the FCC and deallocate messages will need to be
 * set to be equal to the CID allocated at runtime.  Note the CID_POS defines, which indicate the position of the CID in the 
 * message.  Also note that the messages should *not* be consts, since we need to modify them. */

#define QMI_MSG_ALLOCATE_CID_SIZE 16
#define QMI_MSG_FCC_SIZE 13
#define QMI_MSG_DEALLOCATE_CID_SIZE 17

#define QMI_MSG_FCC_CID_POS 5
#define QMI_MSG_DEALLOCATE_CID_POS 16

static guint8 qmi_msg_allocate_cid[QMI_MSG_ALLOCATE_CID_SIZE]     = { 0x01, 0x0F, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x22, 0x00, 0x04, 0x00, 0x01, 0x01, 0x00, 0x02 };
static guint8 qmi_msg_fcc[QMI_MSG_FCC_SIZE]            	          = { 0x01, 0x0C, 0x00, 0x00, 0x02, 0xFF, 0x00, 0x01, 0x00, 0x5F, 0x55, 0x00, 0x00 };
static guint8 qmi_msg_deallocate_cid[QMI_MSG_DEALLOCATE_CID_SIZE] = { 0x01, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x23, 0x00, 0x05, 0x00, 0x01, 0x02, 0x00, 0x02, 0xFF };

/* definitions of the various callbacks we'll need */
static void device_new_ready (GObject *unused, GAsyncResult *res);
static void device_open_ready (MbimDevice *dev, GAsyncResult *res);
static void query_qmi_msg_allocate_cid_ready (MbimDevice *dev, GAsyncResult *res);
static void query_qmi_msg_fcc_ready (MbimDevice *dev, GAsyncResult *res);
static void query_qmi_msg_deallocate_ready (MbimDevice *dev, GAsyncResult *res);
static void device_close_ready (MbimDevice *dev, GAsyncResult *res);

/* definitions of utilities */
static void print_usage ();
static void log_handler (const gchar *log_domain, GLogLevelFlags log_level, const gchar *message, gpointer user_data);

int main(int argc, char* argv[])
{
	verbose = TRUE;

	if(argc <= 1) {
	  print_usage();
	  return(1);
	}

	gchar *device_str = (gchar *) argv[1];
	
	if(verbose) {
	    g_log_set_handler (NULL, G_LOG_LEVEL_MASK, log_handler, NULL);
	    g_log_set_handler ("Mbim", G_LOG_LEVEL_MASK, log_handler, NULL);
	    mbim_utils_set_traces_enabled (TRUE);
	}

	GFile *file;
	cancellable = g_cancellable_new();

	file = g_file_new_for_commandline_arg(device_str);
	loop = g_main_loop_new(NULL, FALSE);
	
	/* the order of asynchronous events is: 
	 *    1) request a new MBIM device,
	 *    2) open the MBIM device,
	 *    3) send a QMI CTL allocate message,
	 *    4) send a QMI FCC authorized message (0x555F),
	 *    5) send a QMI CTL deallocate message,
	 *    6) close the MBIM device.
	 */
	
	/* asynchronously request a new device handler */
	mbim_device_new(file, cancellable, (GAsyncReadyCallback)device_new_ready, NULL);

	g_main_loop_run(loop);

	if (cancellable)
		g_object_unref(cancellable);
	if (device)
		g_object_unref(device);
	g_main_loop_unref(loop);
	g_object_unref(file);

	return(0);
}

static void
device_new_ready (GObject      *unused,
                  GAsyncResult *res)
{
	GError *error = NULL;
	MbimDeviceOpenFlags open_flags = MBIM_DEVICE_OPEN_FLAGS_NONE;

	device = mbim_device_new_finish (res, &error);
	if (!device) {
		g_printerr("error: couldn't create MbimDevice: %s\n",
				error->message);
		exit(EXIT_FAILURE);
	}
	
	mbim_device_open_full(device,
		open_flags,
		30,
		cancellable,
		(GAsyncReadyCallback)device_open_ready,
		NULL);
}

static void
device_open_ready (MbimDevice   *dev,
                   GAsyncResult *res)
{
	GError *error = NULL;

	if (!mbim_device_open_finish (dev, res, &error)) {
		g_printerr ("error: couldn't open the MbimDevice: %s\n",
			error->message);
		exit (EXIT_FAILURE);
	}

	g_debug ("MBIM Device at '%s' ready",
			mbim_device_get_path_display (dev));
	
	/* here we start the QMI stuff... first, get allocate a CID */
	MbimMessage *request;

        g_debug ("Asynchronously sending QMI allocate CID message...");
	//request = (mbim_message_radio_state_query_new (NULL));
        request = (mbim_message_qmi_msg_set_new (QMI_MSG_ALLOCATE_CID_SIZE, ((guint8 *) qmi_msg_allocate_cid), &error));
        mbim_device_command (dev,
                             request,
                             10,
                             cancellable,
                             (GAsyncReadyCallback)query_qmi_msg_allocate_cid_ready,
                             NULL);
	
        mbim_message_unref (request);
        return;
}

static void
query_qmi_msg_allocate_cid_ready (MbimDevice   *dev,
				  GAsyncResult *res)
{
	/* this function will be called when libmbim recieves a message 
	 * 	back from the modem -- if successful, the modem will
	 * 	have given us a CID */
	
	/* now we need to parse the CID that was allocated to us */
	
	/* so, first get the MBIM response */
	MbimMessage *response;
	GError *error = NULL;
	MbimDeviceType device_type;

	response = mbim_device_command_finish (dev, res, &error);
	if (!response || !mbim_message_response_get_result (response, MBIM_MESSAGE_TYPE_COMMAND_DONE, &error)) {
		g_printerr ("error: operation failed: %s\n", error->message);
		g_error_free (error);
		if (response)
			mbim_message_unref (response);
		shutdown (FALSE);
		return;
	}

	/* assuming no failures, we now get the raw data (just a byte array) */
	//guint32 *rawmsglen = (guint32 *) 72;
	guint32 rawmsglen = 72;
	const guint8 *rawmsg = mbim_message_get_raw(response, &rawmsglen, &error);
	qmi_cid = rawmsg[71]; /* 71 because the message is 72 bytes long, and the CID is in the last position */

	g_debug("Allocated CID is: %02X\n", qmi_cid);
	
	
	/* now we modify the FCC message with the new CID */
	qmi_msg_fcc[QMI_MSG_FCC_CID_POS] = qmi_cid;
	
	/* and now we send the FCC authorization message! */
	/* this should bring the modem out of LPM and into life */
	MbimMessage *request;
	
        g_debug ("Asynchronously sending QMI FCC authorization message...");
        request = (mbim_message_qmi_msg_set_new (QMI_MSG_FCC_SIZE, ((guint8 *) qmi_msg_fcc), &error));
	mbim_device_command (dev,
                             request,
                             10,
                             cancellable,
                             (GAsyncReadyCallback)query_qmi_msg_fcc_ready,
                             NULL);
	
        mbim_message_unref (request);
	mbim_message_unref (response);
	return;
}

static void
query_qmi_msg_fcc_ready (MbimDevice   *dev,
			 GAsyncResult *res)
{
	/* if everything went well, the FCC authorization should be set now
	 * 	if we were lazy, we could quit here, but it would be better to 
		deallocate the CID... */
  
	/* let's check for an error first */
	MbimMessage *response;
	GError *error = NULL;
	MbimDeviceType device_type;

	response = mbim_device_command_finish (dev, res, &error);
	if (!response || !mbim_message_response_get_result (response, MBIM_MESSAGE_TYPE_COMMAND_DONE, &error)) {
		g_printerr ("error: operation failed: %s\n", error->message);
		g_error_free (error);
		if (response)
			mbim_message_unref (response);
		shutdown (FALSE);
		return;
	}

	/* assuming no failures, we modify the deallocate message with the CID */
	qmi_msg_deallocate_cid[QMI_MSG_DEALLOCATE_CID_POS] = qmi_cid;
	
	/* and now we send the FCC authorization message! */
	/* this should bring the modem out of LPM and into life */
	MbimMessage *request;
	
        g_debug ("Asynchronously sending QMI deallocate message...");
        request = (mbim_message_qmi_msg_set_new (QMI_MSG_DEALLOCATE_CID_SIZE, ((guint8 *) qmi_msg_deallocate_cid), &error));
	mbim_device_command (dev,
                             request,
                             10,
                             cancellable,
                             (GAsyncReadyCallback)query_qmi_msg_deallocate_ready,
                             NULL);
	
        mbim_message_unref (request);
	mbim_message_unref (response);
	return;
}

static void
query_qmi_msg_deallocate_ready (MbimDevice   *dev,
				GAsyncResult *res)
{
	/* time to start wrapping up, let's deallocate the CID */
  
	/* check for an error first */
	MbimMessage *response;
	GError *error = NULL;
	MbimDeviceType device_type;

	response = mbim_device_command_finish (dev, res, &error);
	if (!response || !mbim_message_response_get_result (response, MBIM_MESSAGE_TYPE_COMMAND_DONE, &error)) {
		g_printerr ("error: operation failed: %s\n", error->message);
		g_error_free (error);
		if (response)
			mbim_message_unref (response);
		shutdown (FALSE);
		return;
	}
	
	/* and now we close the device */
	mbim_device_close (device,
			   15,
			   cancellable,
			   (GAsyncReadyCallback) device_close_ready,
			   NULL);
	
	mbim_message_unref (response);
	return;
}

static void
device_close_ready (MbimDevice   *dev,
                    GAsyncResult *res)
{
    GError *error = NULL;

    if (!mbim_device_close_finish (dev, res, &error)) {
        g_printerr ("error: couldn't close device: %s", error->message);
        g_error_free (error);
    } else
        g_debug ("Device closed");

    g_main_loop_quit (loop);
}

/****** UTILITIES *******/

static void
print_usage()
{
    g_printf("lpmutil -- moves non-QMI-enabled modems out of LPM\n");
    g_printf("              by using the MBIM-over-QMI service\n");
    g_printf("\n");
    g_printf("Usage:\n");
    g_printf("  lpmutil [device]\n");
    g_printf("\n");
    g_printf("Example:\n");
    g_printf("  lpmutil /dev/cdc-wdm0\n");
}

static void
log_handler (const gchar *log_domain,
             GLogLevelFlags log_level,
             const gchar *message,
             gpointer user_data)
{
    const gchar *log_level_str;
    time_t now;
    gchar time_str[64];
    struct tm *local_time;
    gboolean err;

    /* Nothing to do if we're silent */
    //if (silent_flag)
    //    return;

    now = time ((time_t *) NULL);
    local_time = localtime (&now);
    strftime (time_str, 64, "%d %b %Y, %H:%M:%S", local_time);
    err = FALSE;

    switch (log_level) {
    case G_LOG_LEVEL_WARNING:
        log_level_str = "-Warning **";
        err = TRUE;
        break;

    case G_LOG_LEVEL_CRITICAL:
    case G_LOG_FLAG_FATAL:
    case G_LOG_LEVEL_ERROR:
        log_level_str = "-Error **";
        err = TRUE;
        break;

    case G_LOG_LEVEL_DEBUG:
        log_level_str = "[Debug]";
        break;

    default:
        log_level_str = "";
        break;
    }

    if (!verbose && !err)
        return;

    g_fprintf (err ? stderr : stdout,
               "[%s] %s %s\n",
               time_str,
               log_level_str,
               message);
}

